import java.util.Iterator;
import java.util.Vector;
import lotus.domino.*;
import java.io.*;

public class DatabaseNameExtractor {
	String dxl;
	String dxl_original;
	Session session;
	
	class Element {
		public int index;
		public String name;		
		
		public Element(int index, String name){
			this.index = index;
			this.name = name;
		}
	}
	
	public DatabaseNameExtractor(String dxl, Session session) {
		this.dxl_original = dxl;;
		this.dxl = dxl_original.toLowerCase();
		this.session = session;
		
		
		try {
			FileWriter out = new FileWriter(new File("dxl.xml"));
			out.write(dxl_original);
			out.close();
		}
		catch(IOException e){
			e.printStackTrace();
		}
	}
	
	private Element getLastElement(String[] elements, int startIndex){
		String element = "";
		int max = -1;
		int last;
		for (int i=0; i<elements.length; i++){
			last = dxl.lastIndexOf(elements[i], startIndex);
			if (last>max) {
				max = last;
				element = elements[i];
			}
		}
		return new Element(max, element);
	}
	
	private Element getNextElement(String[] elements, int startIndex){
		String element = "";
		int min = Integer.MAX_VALUE;
		int next;
		for (int i=0; i<elements.length; i++){
			next = dxl.indexOf(elements[i], startIndex);
			if (next<min) {
				min = next;
				element = elements[i];
			}
		}
		return new Element(min, element);
	}
	
	private String getAttribute(String name, int startIndex){
		int end = dxl.indexOf(">", startIndex);
		int name_attr_index = dxl.indexOf(name+"='", startIndex);
		int name_attr_end_index = dxl.indexOf("'", name_attr_index+name.length()+2);
		
		if ((name_attr_index>0)&&(name_attr_index<end))
			return dxl_original.substring(name_attr_index+name.length()+2, name_attr_end_index);
		else
			return "";
	}
	
	private String[] getParameters(int startIndex, char separator){
		//String debug = dxl_original.substring(startIndex);
		
		int start = dxl.indexOf("(", startIndex);

		int index = start+1;
		boolean cont = true;
		boolean insideQuota = false;
		int brackets = 0;
		String t = "";
		Vector params = new Vector();
		char c;
		
		while (cont){
			
			c = dxl_original.charAt(index);
			
			if (c=='"') insideQuota = !insideQuota;			
			
			if (c=='(') brackets += 1;			
			
			if (c=='\n') cont = false;
			if ((c==')')&(!insideQuota)&(brackets<1)) cont = false;
			
			if (((c==separator)&(!insideQuota)&(brackets<1))|(!cont)){
				params.add(t);
				t = "";
			}
			else t += c;
			
			if (c==')') brackets -= 1;
			
			index++;
		}
								
		String s2[] = (String[])params.toArray(new String[params.size()]);
		for (int i=0; i<s2.length; i++) s2[i] = s2[i].replace("\"", "").trim();
		return s2;
	}
	
	private int getCodeLineNo(int startIndex){
		String[] prev_elements = {"<code", "<lotusscript", "<java"};
		Element prev_element = getLastElement(prev_elements, startIndex);
		
		int line = -1;
		if ((prev_element.name.equals(prev_elements[1]))|(prev_element.name.equals(prev_elements[2]))){
			int codeStart = dxl.indexOf(">", prev_element.index) + 1;
			String code = dxl_original.substring(codeStart, startIndex);
			
			String[] lines = code.split("\n");
			return lines.length;
			
		}
		
		return line;
	}
	
	private String getParent(int startIndex){
		String parent_str = "";
		int ls_tag_index = dxl.lastIndexOf("<code", startIndex);
											
		String[] elements = {"<action","<agent","<scriptlibrary","<button","<field","<computedtext"};
		Element last_element = getLastElement(elements, ls_tag_index);
		
		if (last_element.name.equals(elements[0])) { //jsme v tagu action
			
			int title_attr_index = dxl.indexOf("title='", last_element.index);
			int title_attr_end_index = dxl.indexOf("'", title_attr_index+7);
			
			parent_str = "Action " + dxl_original.substring(title_attr_index+6, title_attr_end_index+1);
			
			String[] prev_elements = {"<sharedactions", "<subform", "<form ", "<page"};
			Element prev_element = getLastElement(prev_elements, last_element.index);					
			
			if (prev_element.name.equals(prev_elements[1])){ //subform
				parent_str = "Subform '" + getAttribute("name", prev_element.index) + "': " + parent_str;
			}
			else if (prev_element.name.equals(prev_elements[2])){ //form							
				parent_str = "Form '" + getAttribute("name", prev_element.index) + "': " + parent_str;
			}
			else if (prev_element.name.equals(prev_elements[3])){ //page							
				parent_str = "Page '" + getAttribute("name", prev_element.index) + "': " + parent_str;
			} 
			
		}
		else if (last_element.name.equals(elements[1])){ //jsme v tagu agent
			parent_str = "Agent '" + getAttribute("name", last_element.index) + "'";		
		}
		else if (last_element.name.equals(elements[2])){ //jsme v tagu scriptlibrary
			parent_str = "Library '" + getAttribute("name", last_element.index) + "'";
			parent_str = parent_str + ": '" + getAttribute("event", ls_tag_index) + "'";	
		}
		else if (last_element.name.equals(elements[3])){ //jsme v tagu button
			
			String[] prev_elements = {"<subform", "<form ", "<page"};
			Element prev_element = getLastElement(prev_elements, last_element.index);					
			
			if (prev_element.name.equals(prev_elements[0])){ //subform
				parent_str = "Subform '" + getAttribute("name", prev_element.index) + "': " + parent_str;
			}
			else if (prev_element.name.equals(prev_elements[1])){ //form							
				parent_str = "Form '" + getAttribute("name", prev_element.index) + "': " + parent_str;
			}
			else if (prev_element.name.equals(prev_elements[2])){ //page							
				parent_str = "Page '" + getAttribute("name", prev_element.index) + "': " + parent_str;
			} 
						
			int name_attr_index = dxl.indexOf("</code>", last_element.index);
			int name_attr_end_index = dxl.indexOf("</button>", name_attr_index+7);
			parent_str += "Button '" + dxl_original.substring(name_attr_index+7, name_attr_end_index) + "\'";		
		}
		else if (last_element.name.equals(elements[4])){ //jsme v tagu field
			
			String[] prev_elements = {"<form ", "<sharedfield"};
			Element prev_element = getLastElement(prev_elements, last_element.index);
			
			//int form_tag_index = dxl.lastIndexOf("<form ", last_element.index);
			
			if (prev_element.name.equals(prev_elements[0]))
				parent_str = "Form '" + getAttribute("name", prev_element.index) + "': ";
			
			parent_str += "Field '" + getAttribute("name", last_element.index) + "'";
		}
		else if (last_element.name.equals(elements[5])){ //jsme v tagu computedtext
			
			String[] prev_elements = {"<subform", "<form ", "<page"};
			Element prev_element = getLastElement(prev_elements, last_element.index);					
			
			if (prev_element.name.equals(prev_elements[0])){ //subform
				parent_str = "Subform '" + getAttribute("name", prev_element.index) + "': " + parent_str;
			}
			else if (prev_element.name.equals(prev_elements[1])){ //form							
				parent_str = "Form '" + getAttribute("name", prev_element.index) + "': " + parent_str;
			}
			else if (prev_element.name.equals(prev_elements[2])){ //page							
				parent_str = "Page '" + getAttribute("name", prev_element.index) + "': " + parent_str;
			} 
			
			parent_str += "Computed text";
		}
		
		
		//int line = getCodeLineNo(startIndex);
		//if (line>-1) parent_str += ", line " + line; 
		
		return parent_str;
	}
	
	private String getVariableName(int startIndex){							
		int startLineIndex = dxl.lastIndexOf("\n", startIndex);
		
		//if there is '=' then copy everything before
		int eq_sign = dxl.lastIndexOf("=", startIndex); 
		if (eq_sign>startLineIndex) startIndex = eq_sign; 
				
		String s = dxl.substring(startLineIndex+1, startIndex);
				
		String s2 = s.replace("set ", "").replace("dim ", "").replace("as ", "").replace("=", "").trim();
		
		int space = s2.indexOf(" ");
		if (space>0){
			s2 = s2.substring(space+1);
		}
		
		return s2.trim();		
	}
	
	private int getLanguage(int startIndex){
		int codeStartIndex = dxl.lastIndexOf("<code", startIndex);
		
		String[] elements = {"<lotusscript","<java ","<formula"};
		Element last_element = getLastElement(elements, startIndex);		
		
		if (last_element.index<codeStartIndex)
			return 0;
		else {
			if (last_element.name.equals(elements[0])) return 1;
			else if (last_element.name.equals(elements[1])) return 2;
			else return 3;
		}
	}
	
	private Vector getTargetViews(String variable_name, int startIndex){
		String target_str = "";
		Vector targets = new Vector();
		int codeEndIndex = dxl.indexOf("</code>", startIndex);
		int getViewIndex = dxl.indexOf(variable_name + ".getview(", startIndex);
		while ((getViewIndex>=0)&(getViewIndex<codeEndIndex)){		
			
			if (!isComment(getViewIndex)){
			
				char c = dxl.charAt(getViewIndex-1); 
				if ((c==' ')|(c=='=')){
					int viewNameIndex = dxl.indexOf("(", getViewIndex);
					int viewName_end_Index = dxl.indexOf(")", viewNameIndex);
					
					int firstQuota = dxl.indexOf("\"", viewNameIndex);
					if ((firstQuota>0)&(firstQuota<viewName_end_Index)){ //inside viewName is quotation mark
						
						int nextQuota = dxl.indexOf("\"", firstQuota+1);
						
						if ((nextQuota>0)&(nextQuota>viewName_end_Index)) //right bracket before quotation end
							viewName_end_Index = dxl.indexOf(")", viewName_end_Index+1); //goto next bracket
					}										
					
					target_str = dxl_original.substring(viewNameIndex+1, viewName_end_Index);
					if (!(target_str.equals(""))) target_str = "View " + target_str;										
					targets.add(new Element(getCodeLineNo(viewNameIndex), target_str));
				}
			
			}
			getViewIndex = dxl.indexOf(variable_name + ".getview(", getViewIndex+1);
		}
		
		getViewIndex = dxl.indexOf(variable_name + ".getagent(", startIndex);
		while ((getViewIndex>=0)&(getViewIndex<codeEndIndex)){		
			
			if (!isComment(getViewIndex)){
			
				char c = dxl.charAt(getViewIndex-1); 
				if ((c==' ')|(c=='=')){
					int viewNameIndex = dxl.indexOf("(", getViewIndex);
					int viewName_end_Index = dxl.indexOf(")", viewNameIndex);
					
					int firstQuota = dxl.indexOf("\"", viewNameIndex);
					if ((firstQuota>0)&(firstQuota<viewName_end_Index)){ //inside viewName is quotation mark
						
						int nextQuota = dxl.indexOf("\"", firstQuota+1);
						
						if ((nextQuota>0)&(nextQuota>viewName_end_Index)) //right bracket before quotation end
							viewName_end_Index = dxl.indexOf(")", viewName_end_Index+1); //goto next bracket
					}											
					
					target_str = dxl_original.substring(viewNameIndex+1, viewName_end_Index);
					if (!(target_str.equals(""))) target_str = "Agent " + target_str;										
					targets.add(new Element(getCodeLineNo(viewNameIndex), target_str));
				}
			
			}
			getViewIndex = dxl.indexOf(variable_name + ".getagent(", getViewIndex+1);
		}
		
		return targets;
	}
	
	public void getFormulaRelations(Vector relations) {
		//@DbColumn( class : cache ; 		server : database ; view ; columnNumber )
		//@DbLookup( class : cache ; 		server : database ; view ; key ; fieldName ; keywords )
		//@Command( [FileOpenDatabase]; 	server : database ; viewName ; key ; newinstance ; temporary )
		//@Command([Compose]; 				server : database ; form ; width : height) 
		//@Command([ComposeWithReference]; 	server : database ; form ; flags) 
		
		
		String[] search_strings = {"@dbcolumn(", "@dblookup(", "@command("}; 
				
		for (int i=0; i<search_strings.length; i++){
			
			int occurenceIndex = dxl.indexOf(search_strings[i]);		
				
			while (occurenceIndex>=0)  {
												
				//String debug = dxl_original.substring(occurenceIndex);
				
				String[] params = getParameters(occurenceIndex, ';');
				
				if (params.length>0) {

					if ((i!=3)|(params[0].equals("[FileOpenDatabase]")|params[0].equals("[Compose]")|params[0].equals("[ComposeWithReference]"))){

						String server = "";
						String db_name = "";
						String view_name = "";

						if (params.length>2){
							db_name = params[1];
							if (params[1].contains(":")){
								int comma = params[1].indexOf(":");
								server = params[1].substring(0, comma);
								db_name = params[1].substring(comma+1);
							}

							view_name = "View ";				
							if (params[0].equals("[Compose]")|params[0].equals("[ComposeWithReference]")) view_name = "Form ";				
							view_name += params[2];
						}

						if (!db_name.equals("")){
							Relation rel = new Relation(getParent(occurenceIndex), server, db_name, view_name, 3, occurenceIndex);
							resolveDbTitle(rel);
							relations.add(rel);
						}
					}
				}

				occurenceIndex = dxl.indexOf(search_strings[i], occurenceIndex+1);
				
			}
		}
	}
	
	public String getBase64(){		
		int dataStart = dxl.indexOf("IconBitmap");
		dataStart = dxl.indexOf("<rawitemdata type='6'>", dataStart);
		int dataEnd = dxl.indexOf("</rawitemdata>");
		return dxl_original.substring(dataStart+22, dataEnd).trim();
	}
	
	public void resolveDbTitle(Relation relation){
		String db_name = relation.target_database;
		String title = db_name;
		if ((db_name.length()==16)&&(db_name.indexOf(".nsf")==-1)){ //16 chars and not .nsf
			try {
				DbDirectory dir = session.getDbDirectory(null);
				Database db = dir.openDatabaseByReplicaID(db_name);
				if (db!=null){
					if (!db.getTitle().equals("")){
						title = db.getTitle();
						relation.setTargetDbFictional(false);
					}
				}
			}
			catch (Exception e){}
		}
		else {
			if (db_name.indexOf(".nsf")!=-1){
				try {
					Database db = session.getDatabase(session.getServerName(), db_name);
					if (db!=null){
						if (!db.getTitle().equals("")){
							title = db.getTitle();
							relation.setTargetDbFictional(false);
						}
					}
				}
				catch (Exception e){}
			}
		}
		relation.target_database = title;
	}
	
	public boolean isComment(int startIndex){
		
		//single line comments
		int lineEnd = dxl.lastIndexOf("\n", startIndex);
		int ls_comment = dxl.lastIndexOf("\'", startIndex);
		int java_comment = dxl.lastIndexOf("//", startIndex);
		
		if ((ls_comment>lineEnd)|(java_comment>lineEnd)){
			return true;
		}
		else {
			
			//multi line LS comments
			
			String[] elements = {"<code","%end rem","%rem"};
			Element last_element = getLastElement(elements, startIndex);
			
			if (last_element.name.equals(elements[2])){ //%REM
				return true;
			}
			else {
				
				//multi line Java comments
				
				String[] elements2 = {"<code","/*","*/"};
				Element last_element2 = getLastElement(elements2, startIndex);
				
				if (last_element2.name.equals(elements2[1])){
					return true;
				}
				else return false;				
			}			
		}		
	}

	public void getNewDatabaseRelations(Vector relations) {
		
		String search_string = "new notesdatabase(";
		String variable_name = "";
		
		int occurenceIndex = dxl.indexOf(search_string);		
		
		while (occurenceIndex>=0)  {
			
			if (!isComment(occurenceIndex)){
			
				variable_name = getVariableName(occurenceIndex);

				
				String[] params = getParameters(occurenceIndex, ',');				
				if (params.length>1) {
				
				/*
				int paramStartIndex = dxl.indexOf("(", occurenceIndex);
				int commaIndex = dxl.indexOf(",", occurenceIndex);
				int endIndex = dxl.indexOf(")", paramStartIndex); 
				String server = dxl_original.substring(paramStartIndex+1, commaIndex).replace("\"", "").trim(); 
				String db_name = dxl_original.substring(commaIndex+1, endIndex).replace("\"", "").trim();
				*/

					if (!params[0].equals("")){

						Vector targets = getTargetViews(variable_name, occurenceIndex);
						Iterator it = targets.iterator ();
						while (it.hasNext ()) {
							Element target = (Element) it.next ();
							Relation rel = new Relation(getParent(occurenceIndex) + ", line " + target.index, params[0], params[1], target.name, getLanguage(occurenceIndex), occurenceIndex);
							resolveDbTitle(rel);
							relations.add(rel);
						}

						if (targets.size()==0){
							Relation rel = new Relation(getParent(occurenceIndex)+ ", line " + getCodeLineNo(occurenceIndex), params[0], params[1], "", getLanguage(occurenceIndex), occurenceIndex);
							resolveDbTitle(rel);
							relations.add(rel);
						}
				
					}
				
				}

			}
	        
			occurenceIndex = dxl.indexOf(search_string, occurenceIndex+1);
		}
				
	}		
	
	public void getSessionGetDatabaseRelations(Vector relations) {
		
		String search_string = ".getdatabase(";
		String variable_name = "";
		
		int occurenceIndex = dxl.indexOf(search_string);
		
		while (occurenceIndex>=0)  {
			
			if (!isComment(occurenceIndex)){
			
				variable_name = getVariableName(occurenceIndex);

				String params[] = getParameters(occurenceIndex, ',');				
				
				if (params.length>1){
				
					String server = params[0]; 
					String db_name = params[1];

					Vector targets = getTargetViews(variable_name, occurenceIndex);
					Iterator it = targets.iterator ();
					while (it.hasNext ()) {
						Element target = (Element) it.next ();
						Relation rel = new Relation(getParent(occurenceIndex)+ ", line " + target.index, server, db_name, target.name, getLanguage(occurenceIndex), occurenceIndex);
						resolveDbTitle(rel);
						relations.add(rel);
					}

					if (targets.size()==0){
						Relation rel = new Relation(getParent(occurenceIndex)+ ", line " + getCodeLineNo(occurenceIndex), server, db_name, "", getLanguage(occurenceIndex), occurenceIndex);
						resolveDbTitle(rel);
						relations.add(rel);
					}
				}
			}

			occurenceIndex = dxl.indexOf(search_string, occurenceIndex+1);
		}
						
	}
	
	public void getOpenDatabaseRelations(Vector relations) {
		
		String search_string = ".opendatabase(";
		String variable_name = "";
		
		int occurenceIndex = dxl.indexOf(search_string);
		
		while (occurenceIndex>=0)  {
			
			if (!isComment(occurenceIndex)){
														
				String[] params = getParameters(occurenceIndex,',');
				
				String server = "";
				if (params.length>0) server = params[0];
				
				String db_name = "";
				if (params.length>1) db_name = params[1];
				
				String target = "";
				if (params.length>2) target = "View " + params[2];
				
				if (params.length>0){
					Relation rel = new Relation(getParent(occurenceIndex)+ ", line " + getCodeLineNo(occurenceIndex), server, db_name, target, getLanguage(occurenceIndex), occurenceIndex);
					resolveDbTitle(rel);
					relations.add(rel);	 
				}
			}

			occurenceIndex = dxl.indexOf(search_string, occurenceIndex+1);
		}
						
	}
	
public void getOpenByReplicaID(Vector relations) {
		
		String search_string = ".openbyreplicaid(";
		
		int occurenceIndex = dxl.indexOf(search_string);
		
		while (occurenceIndex>=0)  {
			
			if (!isComment(occurenceIndex)){
										
				String[] params = getParameters(occurenceIndex, ',');
				
				String replicaID = "";
				String server = "";
				if (params.length>1) {
					server = params[1];
					replicaID = params[1];				
					Relation rel = new Relation(getParent(occurenceIndex)+ ", line " + getCodeLineNo(occurenceIndex), server, replicaID, "", getLanguage(occurenceIndex), occurenceIndex);
					resolveDbTitle(rel);
					relations.add(rel);	 
				}
			}

			occurenceIndex = dxl.indexOf(search_string, occurenceIndex+1);
		}
						
	}

	
	public void getFrameRelations(Vector relations) {
		
		String search_string = "<namedelementlink";		
		
		int occurenceIndex = dxl.indexOf(search_string);		
		
		while (occurenceIndex>=0)  {	
			
			//String debug = dxl_original.substring(occurenceIndex, occurenceIndex+100);
			
			String[] elements = {"<frameset"};
			Element parent_frameset = getLastElement(elements, occurenceIndex);
			String parent_frameset_name = getAttribute("name", parent_frameset.index);
			
			String type = getAttribute("type", occurenceIndex);						
			String server = ""; 
			String db_name = getAttribute("database", occurenceIndex);
			String target = type + " '" + getAttribute("name", occurenceIndex) + "'";
						
			if (db_name.equals("")){
				int frame_end = dxl.indexOf("</frame>", occurenceIndex);												
				int formula_start = dxl.indexOf("<formula", occurenceIndex);
				
				if ((formula_start>0)&&(frame_end>0)&&(formula_start<frame_end)){
				
					int formula_end = dxl.indexOf("</formula", formula_start);
					target = dxl_original.substring(formula_start+9, formula_end).replace("\"", "");

					formula_start = dxl.indexOf("<formula", formula_end);
					formula_end = dxl.indexOf("</formula", formula_start);											

					type = dxl_original.substring(formula_start+9, formula_end).replace("\"", "");

					target = type + " '" + target + "'";

					formula_start = dxl.indexOf("<formula", formula_end);
					formula_end = dxl.indexOf("</formula", formula_start);							

					db_name = dxl_original.substring(formula_start+10, formula_end);
				}
			}		
			
			if (!db_name.equals("")){
				Relation rel = new Relation("Frameset '"+parent_frameset_name+"'", server, db_name, target, 0, occurenceIndex);
				resolveDbTitle(rel);
				relations.add(rel);
			}
	        	        	        
			occurenceIndex = dxl.indexOf(search_string, occurenceIndex+1);
		}
				
	}	
	
	public void getOutlineRelations(Vector relations, Database db) {
		
		Outline outline;
		
		String search_string = "<item name='$title'><text>";		
		
		int occurenceIndex = dxl.indexOf(search_string);		
		
		while (occurenceIndex>=0)  {
			int item_end = dxl.indexOf("</text>", occurenceIndex);
			
			String outline_name = dxl_original.substring(occurenceIndex+26, item_end);
			
			try {
				outline = db.getOutline(outline_name);

				if (outline!=null){
					OutlineEntry entry = outline.getFirst();
				    while (entry != null) {
				    	
				    	//if (!entry.isInThisDB()){				    						    						    			
				    			String server = "<Unknown>"; 
				    			String db_name = "<Unknown>";
				    			
				    			Database target_db = entry.getDatabase();
				    			if (target_db!=null){
				    				server = target_db.getServer(); 
					    			db_name = target_db.getFilePath();
				    			}
				    			
				    			String s1 = entry.getFormula();
				    			if (s1.equals("")) s1 = entry.getNamedElement();
				    			
				    			String entryClass = "";
				    	        switch (entry.getEntryClass()) {
				    	          case OutlineEntry.OUTLINE_CLASS_DATABASE:
				    	            entryClass = "Database"; break;
				    	          case OutlineEntry.OUTLINE_CLASS_DOCUMENT:
				    	            entryClass = "Document"; break;
				    	          case OutlineEntry.OUTLINE_CLASS_FORM:
				    	            entryClass = "Form"; break;
				    	          case OutlineEntry.OUTLINE_CLASS_FOLDER:
				    	            entryClass = "Folder"; break;
				    	          case OutlineEntry.OUTLINE_CLASS_FRAMESET:
				    	            entryClass = "Frame set"; break;
				    	          case OutlineEntry.OUTLINE_CLASS_NAVIGATOR:
				    	            entryClass = "Navigator"; break;
				    	          case OutlineEntry.OUTLINE_CLASS_PAGE:
				    	            entryClass = "Page"; break;
				    	          case OutlineEntry.OUTLINE_CLASS_UNKNOWN:
				    	            entryClass = "Unknown"; break;
				    	          case OutlineEntry.OUTLINE_CLASS_VIEW:
				    	            entryClass = "View"; break;
				    	          }
				    							    			
				    			Relation rel = new Relation("Outline '"+outline_name+"': entry '"+entry.getLabel()+"'", server, db_name, entryClass + " '" + s1 + "'", 0, occurenceIndex);
				    			resolveDbTitle(rel);
				    			relations.add(rel);
				    	
				    	//}				    	
				    	
				    	entry = outline.getNext(entry);
				    }
				}
			}
			catch (Exception e) {}
			
			
			occurenceIndex = dxl.indexOf(search_string, occurenceIndex+1);
		}
	}
}
